%{
/*  
    VTun - Virtual Tunnel over TCP/IP network.

    Copyright (C) 1998,1999  Maxim Krasnyansky <max_mk@yahoo.com>

    VTun has been derived from VPPP package by Maxim Krasnyansky. 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 */

/*
 * Version: 1.5 05/01/1999 Maxim Krasnyansky <max_mk@yahoo.com>
 */ 

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <syslog.h>
#include <stdarg.h>

#include "vtun.h"

int lineno = 1;

struct vtun_host default_host;
struct vtun_host *parse_host;

char * search_host;

extern struct vtun_host *host_list;

int cfg_error(const char *fmt, ...);

int yyparse(void);
int yylex(void);	
int yyerror(char *s); 

#define YYERROR_VERBOSE 1

%}

%union {
   char *str;
   int  num;
   struct { int num1; int num2; } dnum;
}

%token K_OPTIONS K_DEFAULT K_PORT K_PERSIST K_TIMEOUT K_PASSWD 
%token K_PROG K_ARGS K_SPEED
%token K_TYPE K_PROT K_COMPRESS K_ENCRYPT  

%token <str> K_HOST K_ERROR 

%token <str> WORD IPADDR NMADDR STRING PATH
%token <num> NUM 
%token <dnum> DNUM

%%
config: 
  | config statement
  ;

statement: '\n'
  | K_OPTIONS   '{' options '}' 

  | K_DEFAULT   { 
		  parse_host = &default_host; 
                }        
    '{' host_options '}' 
  | K_HOST      { 
		  /* Is it the host we are looking for */
		  if( search_host ){
		     if( strcmp(search_host, $1) )
			break; 	
		     search_host = (char*) -1;
		  }

		  if( !(parse_host = malloc(sizeof(struct vtun_host))) ){
		     yyerror("No memory for the host");
		     return -1;
		  }

		  /* Fill new host struct with default values.
		   * MUST dup strings to be able to reread config.
		   */
	  	  memcpy(parse_host, &default_host, sizeof(struct vtun_host));
		  parse_host->host = strdup($1);
		  parse_host->passwd = NULL;
		  if(default_host.prog)
		     parse_host->prog = strdup(default_host.prog);
		  if(default_host.args)
		     parse_host->args = strdup(default_host.args);

		  /* Add host to the list */
		  parse_host->next = host_list;
		  host_list = parse_host;
		}    
    '{' host_options '}' 
		{
		  /* Return if searching host is found */		
	          if( search_host == (char*)-1 )
    		     YYACCEPT; 
		}
  | K_ERROR	{
		  cfg_error("Invalid clause '%s'",$1);
		  YYABORT;
		}
  ;

options:
    option
  | options option
  ;

/* Don't override command line options */
option: '\n'
  | K_PORT NUM 		{ 
			  if(vtun.port == -1)
			     vtun.port = $2;
			} 
  | K_PERSIST NUM 	{ 
			  if(vtun.persist == -1) 
			     vtun.persist = $2; 	
			}
  | K_TIMEOUT NUM 	{  
			  if(vtun.timeout == -1)
			     vtun.timeout = $2; 	
			}
  | K_ERROR		{
			  cfg_error("Unknown option '%s'",$1);
			  YYABORT;
			}
  ;

host_options:
    host_option
  | host_options host_option
  ;

/* Host options. Must free strings first, because they 
 * could be strduped from default_host */
host_option: '\n'
  | K_PASSWD WORD 	{ 
			  free(parse_host->passwd);  
			  parse_host->passwd = strdup($2); 	
			}	
  | K_PROG PATH 	{   
			  free(parse_host->prog);  
			  parse_host->prog = strdup($2); 	
			}
  | K_ARGS STRING 	{   
			  free(parse_host->args);  
			  parse_host->args = strdup($2); 	
			}		
  | K_SPEED NUM 	{ 
			  if( $2 ){ 
			     parse_host->spd_in = parse_host->spd_out = $2;
			     parse_host->flags |= VTUN_SHAPE;
			  } else 
			     parse_host->flags &= ~VTUN_SHAPE;
				
			}
  | K_SPEED DNUM 	{ 
			  if( yylval.dnum.num1 || yylval.dnum.num2 ){ 
			     parse_host->spd_out = yylval.dnum.num1;
		             parse_host->spd_in = yylval.dnum.num2; 	
			     parse_host->flags |= VTUN_SHAPE;
			  } else 
			     parse_host->flags &= ~VTUN_SHAPE;
			}
  | K_COMPRESS NUM 	{  
			  if( $2 ) 
			     parse_host->flags |= VTUN_COMPRESS; 
			  else	
			     parse_host->flags &= ~VTUN_COMPRESS; 
			}
  | K_ENCRYPT NUM 	{  
			  if( $2 ) 
			     parse_host->flags |= VTUN_ENCRYPT; 	
			  else	
			     parse_host->flags &= ~VTUN_ENCRYPT; 	
			}
  | K_TYPE NUM 		{  
			  parse_host->flags &= ~VTUN_TYPE_MASK;
			  parse_host->flags |= $2;
			  if( $2 == VTUN_ETHER ){
			     parse_host->flags &= ~VTUN_PROT_MASK;
			     parse_host->flags |= VTUN_UDP;
			  }
			}	
  | K_PROT NUM 		{  
			  parse_host->flags &= ~VTUN_PROT_MASK;
			  parse_host->flags |= $2;
			}
  | K_ERROR		{
			  cfg_error("Unknown option '%s'",$1);
			  YYABORT;
			}
  ;
%%

int cfg_error(const char *fmt, ...)
{
	char buf[255];
	va_list ap;

	/* print the argument string */
	va_start(ap, fmt);
	vsnprintf(buf,sizeof(buf),fmt,ap);
	va_end(ap);

	yyerror(buf);
	return 0;
}

int yyerror(char *s) 
{
   syslog(LOG_ERR, "%s line %d\n", s, lineno);
   return 0;
}

/* Find host:passwd in the host list */ 
struct vtun_host * find_host(char *host,char *pwd)
{
   struct vtun_host *host_ptr = host_list;

   while( host_ptr ){
     if( !strcmp(host_ptr->host,host) && !strcmp(host_ptr->passwd,pwd) )
	return host_ptr;
     host_ptr = host_ptr->next;
   } 
   return NULL;
}

void free_host_list(void)
{
   struct vtun_host *host_ptr, *next = host_list;

   free(default_host.passwd);   
   free(default_host.prog);   
   free(default_host.args);   
   memset(&default_host, 0, sizeof(struct vtun_host));

   while( (host_ptr = next) ){
     free(host_ptr->host);   
     free(host_ptr->passwd);   
     free(host_ptr->prog);   
     free(host_ptr->args);   
     next = host_ptr->next;
     free(host_ptr);
   } 

   host_list = NULL;
}

/* 
 * Read config file. 
 * If host arg is not NULL, only that host will be read.
 */ 
struct vtun_host * read_config(char *file, char *host) 
{
   static int list_filled = 0;
   extern FILE *yyin;

   if( list_filled )
      free_host_list(); 

   list_filled = 1;

   /* Initialize default options */
   default_host.flags = VTUN_TTY | VTUN_TCP; 

   if( !(yyin = fopen(file,"r")) ){
      syslog(LOG_ERR,"Can not open %s", file);
      return NULL;      
   }

   search_host = host;

   yyparse();

   fclose(yyin);

   return host_list;
}
