/*  
    VTun - Virtual Tunnel over TCP/IP network.

    Copyright (C) 1998,1999  Maxim Krasnyansky <max_mk@yahoo.com>

    VTun has been derived from VPPP package by Maxim Krasnyansky. 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 */

/*
 * Version: 1.5 05/01/1999 Maxim Krasnyansky <max_mk@yahoo.com>
 */ 

#include "config.h"

#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <strings.h>
#include <stdarg.h>
#include <errno.h>
#include <sys/time.h>
#include <syslog.h>

#ifdef HAVE_NETINET_IP_H
#include <netinet/ip.h>
#endif

#ifdef HAVE_NETINET_TCP_H
#include <netinet/tcp.h>
#endif

#include "vtun.h"
#include "linkfd.h"
#include "lib.h"

extern int errno;

/* Functions to manipulate with program title */

extern char **environ;
char	*title_start;	/* start of the proc title space */
char	*title_end;     /* end of the proc title space */
int	title_size;

void init_title(int argc,char *argv[], char *envp[], char *name)
{
	int i;

	/*
	**  Move the environment so settitle can use the space at
	**  the top of memory.
	*/

	for (i = 0; envp[i]; i++);

	environ = (char **) malloc(sizeof (char *) * (i + 1));

	for(i = 0; envp[i]; i++)
	   environ[i] = strdup(envp[i]);
	environ[i] = NULL;

	/*
	**  Save start and extent of argv for set_title.
	*/

	title_start = argv[0];

	/*
	**  Determine how much space we can use for set_title.  
	**  Use all contiguous argv and envp pointers starting at argv[0]
 	*/
	for(i=0; i<argc; i++)
	    if( !i || title_end == argv[i])
	       title_end = argv[i] + strlen(argv[i]) + 1;

	for(i=0; envp[i]; i++)
  	    if( title_end == envp[i] )
	       title_end = envp[i] + strlen(envp[i]) + 1;
	
	strcpy(title_start, name);
	title_start += strlen(name);
	title_size = title_end - title_start;
}

void set_title(const char *fmt, ...)
{
	char buf[255];
	va_list ap;

	memset(title_start,0,title_size);

	/* print the argument string */
	va_start(ap, fmt);
	vsprintf(buf, fmt, ap);
	va_end(ap);

	if( strlen(buf) > title_size - 1)
	   buf[title_size - 1] = '\0';

	strcat(title_start, buf);
}

/* 
 * Allocate psevdo tty, returns master side fd. 
 * Stores slave name in the first arg(must be large enought).
 */  
int getpty(char *sl_name)
{
	char ptyname[] = "/dev/ptyXY";
	char ch[] = "pqrstuvwxyz";
	char digit[] = "0123456789abcdef";
	int l, m;
	int mr_fd;
	
	for(l=0; ch[l]; l++ ){
		for(m=0; digit[m]; m++ ){
			ptyname[8] = ch[l];
			ptyname[9] = digit[m];
			/* Open the master */
			if((mr_fd=open(ptyname, O_RDWR)) < 0)
				continue;
			/* Check the slave */
			ptyname[5] = 't';
			if( (access(ptyname, R_OK | W_OK)) < 0 ){
				close(mr_fd);
			 	ptyname[5] = 'p';
				continue;
			}
			strcpy(sl_name,ptyname);
			return mr_fd;
		}
	}
	return -1;
}

#ifdef HAVE_ETHERTAP
/* 
 * Allocate ether TAP device, returns opened fd. 
 * Stores tapdev name in the first arg(must be large enought).
 */  
int gettap(char *dev)
{
	char tapname[12];
	int i, fd;
	
	for(i=0; i < 16; i++){
	   sprintf(tapname,"/dev/tap%d",i);
	   /* Open device */
	   if( (fd=open(tapname, O_RDWR)) > 0 ){
	      sprintf(dev,"tap%d",i);
	      return fd;
	   }
	}
	return -1;
}
#else
int gettap(char *dev)
{
	syslog(LOG_ERR, "Ethernet support not installed");
	return -1;
}
#endif /* ETHERTAP */


/* Read exacly len bytes */
int read_n(int fd, char *buf, int len)
{
	int t=0, w;

	do {
 	  if( (w = read(fd, buf, len)) <= 0 )
	     return w;
	  len -= w;
	  buf += w;
	  t += w;
	} while(len > 0);

	return(t);
}   

/* Write exacly len bytes */
int write_n(int fd, char *buf, int len)
{
	int t=0, w;

	do {
 	  if( (w = write(fd, buf, len)) < 0 )
	     return -1;
	  len -= w;
	  buf += w;
	  t += w;
	} while(len > 0);

	return(t);
}

/* 
 * Print padded messages.
 * used by 'auth' function to force all messages 
 * to be the same len.
 */
int print_p(int fd,const char *fmt, ...)
{
	char buf[VTUN_MESG_SIZE];
	va_list ap;

	memset(buf,0,sizeof(buf));

	/* print the argument string */
	va_start(ap, fmt);
	vsnprintf(buf,sizeof(buf)-1, fmt, ap);
	va_end(ap);
  
	return write(fd, buf, sizeof(buf));
}

/* Connect with timeout */
int connect_t(int s, struct sockaddr *svr, time_t timeout) 
{
	int sock_flags;
	fd_set fdset;
	struct timeval tv;

	tv.tv_usec=0; tv.tv_sec=timeout;

	sock_flags=fcntl(s,F_GETFL);
	if( fcntl(s,F_SETFL,O_NONBLOCK) < 0 )
	  return -1;

	if( connect(s,svr,sizeof(struct sockaddr)) < 0 && errno != EINPROGRESS)
	   return -1;

	FD_ZERO(&fdset);
	FD_SET(s,&fdset);
	if( select(s+1,NULL,&fdset,NULL,&tv) > 0 ){
	   int l=sizeof(errno);	 
 	   errno=0;
 	   getsockopt(s,SOL_SOCKET,SO_ERROR,&errno,&l);
	} else
	   errno=ETIMEDOUT;  	

	fcntl(s,F_SETFL,sock_flags); 

 	if(errno)
	   return -1;

	return 0;
}	

/* Read with timeout */

int read_t(int fd, void *buf, size_t count, time_t timeout) 
{
	fd_set fdset;
	struct timeval tv;

	tv.tv_usec=0; tv.tv_sec=timeout;

	FD_ZERO(&fdset);
	FD_SET(fd,&fdset);
	if( select(fd+1,&fdset,NULL,NULL,&tv) <= 0)
	   return -1;

	return read(fd,buf,count);
}

/* 
 * Establish UDP session with host connected to fd(socket).
 * Returns connected UDP socket or -1 on error. 
 */
int udp_session(int fd, time_t timeout) 
{
	struct sockaddr_in saddr; 
	short  port;
        int   s,opt;

     	if( (s=socket(AF_INET,SOCK_DGRAM,0))== -1 ){
	   syslog(LOG_ERR,"Can't create socket");
	   return -1;
     	}

	memset(&saddr, 0, sizeof(saddr));
	saddr.sin_family = AF_INET;
	saddr.sin_addr.s_addr = INADDR_ANY;
     	saddr.sin_port = 0;
	
     	if( bind(s,(struct sockaddr *)&saddr,sizeof(saddr)) ){
	   syslog(LOG_ERR,"Can't bind to the socket");
   	   return -1;
     	}

	opt = sizeof(saddr);
	if( getsockname(s,(struct sockaddr *)&saddr,&opt) ){
	   syslog(LOG_ERR,"Can't get socket name");
   	   return -1;
     	}
		
	/* Write port no of the new UDP socket */
	port = saddr.sin_port;
	if( write(fd,&port,sizeof(short)) < 0 ){
	   syslog(LOG_ERR,"Can't write port number");
	   return -1;
     	}

	/* Read port no of the other's end UDP socket */
	if( read_t(fd,&port,sizeof(short),timeout) < 0 ){
	   syslog(LOG_ERR,"Can't read port number");
	   return -1;
     	}

	opt = sizeof(saddr);
	if( getpeername(fd,(struct sockaddr *)&saddr,&opt) ){
	   syslog(LOG_ERR,"Can't get peer name");
   	   return -1;
     	}

	saddr.sin_port = port;
	if( connect(s,&saddr,sizeof(saddr)) ){
	   syslog(LOG_ERR,"Can't connect socket");
	   return -1;
     	}

	return s;
}

int start_tunnel(struct vtun_host *host){
     int fd[2],null_fd;
     int pid,opt;
     char buf[12]="";

     switch( host->flags & VTUN_TYPE_MASK ){
	case VTUN_TTY:
	   if( (fd[0]=getpty(buf)) < 0){
	      syslog(LOG_ERR,"Can't allocate psevdo tty");
	      return -1;
           }
	   break;
	case VTUN_PIPE:
	   if( socketpair(AF_UNIX, SOCK_STREAM, 0, fd) < 0){
 	      syslog(LOG_ERR,"Can't create pipe");
   	      return -1;
	   }
	   break;
	case VTUN_ETHER:
	   if( (fd[0]=gettap(buf)) < 0){
 	      syslog(LOG_ERR,"Can't allocate tap device");
   	      return -1;
	   }
	   break;
     }
	
     switch( (pid=fork()) ){
	case -1:
	   syslog(LOG_ERR,"Couldn't fork()");
	   return -1;
	case 0:
     	   switch( host->flags & VTUN_TYPE_MASK ){
	      case VTUN_TTY:
     	         /* Run the program on the pseudo-tty */
	         setsid();
	         /* Open pty slave -- it becomes controlling terminal */
	         if( (fd[1] = open(buf, O_RDWR)) < 0){
	            syslog(LOG_ERR,"Couldn't open slave pty");
	            return -1;
	         }
		 /* Fall through */
	      case VTUN_PIPE:
	         if( (null_fd = open("/dev/null", O_RDWR)) < 0){
	            syslog(LOG_ERR,"Couldn't open /dev/null");
	            return -1;
	         }	
	         close(fd[0]);
	         close(0); dup(fd[1]);
                 close(1); dup(fd[1]);
	         close(fd[1]);

	         /* Route stderr to /dev/null */
	         close(2); dup(null_fd);
	         close(null_fd);
	         break;
	      case VTUN_ETHER:
		 /* Nothing to do for Ethernet */
		 break;
	   }
	   exec_program(host,buf);
	   /* Exec error, exit immideately*/
   	   exit(1);           
	default:
	   switch( host->flags & VTUN_TYPE_MASK ){
	      case VTUN_TTY:
	         set_title("%s tty prog pid <%d>", host->host, pid);
	         break;
	      case VTUN_PIPE:
		 /* Close second end of the pipe */
		 close(fd[1]);
	         set_title("%s pipe prog pid <%d>", host->host, pid);
	  	 break;
	      case VTUN_ETHER:
	         set_title("%s ether %s", host->host, buf);
	   	 break;
     	   }

	   switch( host->flags & VTUN_PROT_MASK ){
	      case VTUN_TCP:
	         opt=1;
	         setsockopt(host->rmt_fd,SOL_SOCKET,SO_KEEPALIVE,&opt,sizeof(opt));

	         opt=1;
	         setsockopt(host->rmt_fd,IPPROTO_TCP,TCP_NODELAY,&opt,sizeof(opt));
	         break;
	      case VTUN_UDP:
	         if( (opt = udp_session(host->rmt_fd,VTUN_TIMEOUT)) == -1){
		    syslog(LOG_ERR,"Can't establish UDP session");
		    return -1;
	         } 	
	         /* UDP session, close TCP socket */	
	         close(host->rmt_fd); host->rmt_fd = opt;		
	  	 break;
     	   }

	   host->loc_fd = fd[0];
	   linkfd(host);

	   close(fd[0]);
	   return 0;
     }
}

/* 
 * Split args string and execute program.
 * Substitutes opt in place off '%%'.
 * Understands single quotes. 
 */ 
void exec_program(struct vtun_host *host, char *opt){
     char *argv[50],*str;
     int i = 1;	

     if( (str=strrchr(host->prog,'/')) )
        argv[0] = str+1;	
     else
        argv[0] = host->prog;	
	
     if(host->args){
        str = host->args;

        while( *str && i < 50){
	  if( *str == ' ' ){
	     str++;
	     continue; 
	  }

	  if( *str == '%' && *(str+1) == '%' ){
	     argv[i++] = strdup(opt);
	     str+=2;	
	     continue;
	  }

	  if( *str == '\'' ){
	     argv[i++] = ++str;
	     while( *str ){
		if( *str == '\'' ){
		   *(str++) = '\0';
		   break;
		}
		str++;
	     }
	     continue;
	  }	     

	  argv[i++] = str;
	  while( *str ){
	     if( *str == ' ' ){
	        *(str++) = '\0';
	        break;
	     }
	     str++;
	  }
        }
     }
     argv[i]=NULL;

     execv(host->prog,argv);

     syslog(LOG_ERR,"Can't exec program %s",host->prog);
}
  
/* 
 * Functions to convert binary flags to character string and back.
 * string format:  <CS64> 
 * C - compression, S - speed for shaper and so on.
 */ 

char *bf2cf(struct vtun_host *host)
{
     static char str[20], *ptr = str;

     *(ptr++) = '<';

     switch( host->flags & VTUN_PROT_MASK ){
	case VTUN_TCP:
	   /* Don't add this flag for now, 
	    * it will break old versions. */
	   break;

	case VTUN_UDP:
	   *(ptr++) = 'U';
	   break;
     }

     switch( host->flags & VTUN_TYPE_MASK ){
	case VTUN_TTY:
	   *(ptr++) = 't'; 	
	   break;

	case VTUN_PIPE:
	   *(ptr++) = 'p';
	   break; 	

	case VTUN_ETHER:
	   *(ptr++) = 'e';
	   break;
     } 

     if( (host->flags & VTUN_SHAPE) && host->spd_in )
	sprintf(ptr,"S%d",host->spd_in);
 
     if( host->flags & VTUN_COMPRESS )
	strcat(ptr,"C"); 	

     if( host->flags & VTUN_ENCRYPT )
	strcat(ptr,"E"); 	

     strcat(ptr,">");

     return str;
}

/* return 1 on success, otherwise 0 */

int cf2bf(char *str, struct vtun_host *host)
{
     char *ptr, *p;
     int s;

     if( (ptr = strchr(str,'<')) ){ 
	ptr++;
	while(*ptr){  
	   switch(*ptr++){
	     case 't':
		host->flags |= VTUN_TTY;
		break;
	     case 'p':
		host->flags |= VTUN_PIPE;
		break;
	     case 'e':
		host->flags |= VTUN_ETHER;
		break;
	     case 'U':
		host->flags &= ~VTUN_PROT_MASK;
		host->flags |= VTUN_UDP;
		break;
	     case 'T':
		host->flags &= ~VTUN_PROT_MASK;
		host->flags |= VTUN_TCP;
		break;
	     case 'C':
		host->flags |= VTUN_COMPRESS;
		break;
	     case 'E':
		host->flags |= VTUN_ENCRYPT;
		break;
     	     case 'S':
		if((s = strtol(ptr,&p,10)) == ERANGE || ptr == p) 
		   return 0;
		if( s ){
	    	   host->flags |= VTUN_SHAPE;
		   host->spd_out = s; 
		}
		ptr = p;
		break;
	     case '>':
	        return 1;
	     default:
		return 0;
	   }
	}
     }
     return 0;
}
